/**** Yet Another Alpha Simulator: SimAlpha        by Kenji KISE ****/
/********************************************************************/
#include "define.h"

/**********************************************************************/
int system_manager::translate_fd(DATA_TYPE fd){
  int fd_t = -1;
  for(int i=0; i<FD_MAPPING_MAX; i++){ 
    if(fd_table[i] == fd) {fd_t = i; break;}
  }
  if(fd_t==-1){printf("*** Error in FD mapping\n"); exit(1);}
  return fd_t;
}

/**********************************************************************/
system_manager::system_manager(architecture_state *as_t, 
			       memory_system *mem_t,
			       struct system_config *ms_t, 
			       struct evaluation_result *e_t){
  as  =  as_t;
  mem = mem_t;
  ms  =  ms_t;
  e   =   e_t;

  uniq     = 0;
  running  = 1;

  for(int i=0; i<FD_MAPPING_MAX; i++) fd_table[i]= i+1;

  /*** This part should be modified for each application. ***/
  /*** fd_table[6]=7; maps real fd(6) to logical fd(7).   ***/
  fd_table[0] = 0; // STDIN
  fd_table[1] = 1; // STDOUT
  fd_table[2] = 2; // STDERR
}

/**********************************************************************/
void system_manager::execute_pal(class instruction *p){
  INST_TYPE ir = p->get_ir();
  int64_t ret=0;
  int F = ms->syscall_flag;

  if((ir & 0xFFFF)!=0x0083){
    switch((ir & 0xFFFF)){
    case 0x009e: // rdunique(Read Unique Value)
      as->r[0].init(uniq);
      if(F) printf("%8qd: call_pal rduniq = %llx\n", 
		   e->retired_inst, uniq);
      return;
    case 0x009f: // wrunique(Write Unique Value)
      uniq  = as->r[16].ld();
      if(F) printf("%8qd: call_pal wrruniq = %llx\n", 
		   e->retired_inst, as->r[16].ld());
      return;
    default:
      if(F){
	printf("*** PAL (ir & 0xFFFF)!=0x0083  Function %x", 
	       (ir & 0xffff));
	printf("  TC:%qd default. %llx\n",
	       e->retired_inst, as->r[0].ld());
      }
      return; 
    }
  }

  if(as->r[0].ld()==0){
    as->r[19].init(0);
    return;
  }


  switch(as->r[0].ld()){
    /************************************************************/    
  case ASYS_UNIMPLEMENT:
    {
      ret = -1;
      as->r[0].init(0);
      as->r[19].init(ret);
      if(F) printf("%8qd: SYS_226 = %lld\n",
		   e->retired_inst, ret);
      break;
    }
    /************************************************************/    
  case ASYS_unlink:
    {
      char buf[512]; /** file name **/
      for(int i=0; i<=512; i++){
	data_t d;
	data_t adr;
	adr.init(as->r[16].ld()+i);
	mem->ld_nbyte(1, &adr, &d);
	buf[i] = d.ld();
	if(buf[i]=='\0') break;
      }
      buf[1]='S';
      buf[2]='A';

      ret = unlink(buf);
      if(ret==-1){
	as->r[0].init(errno);
	as->r[19].init(ret);
      }
      else{
	as->r[19].init(0);
      }

      if(F) printf("%8qd: unlink(%s) = %qd\n",
		   e->retired_inst, buf, ret);
      break;
    }
    /************************************************************/
  case ASYS_open:
    {
      int fd_t;

      char buf[512]; /** file name **/
      for(int i=0; i<=512; i++){
	data_t d;
	data_t adr;
	adr.init(as->r[16].ld()+i);
	mem->ld_nbyte(1, &adr, &d);
	buf[i] = d.ld();
	if(buf[i]=='\0') break;
      }
      //      if(as->r[17].ld()==0){ /** RDONLY **/
      if(as->r[17].ld()==0 || as->r[17].ld()==2){ /** RDONLY or RDWR **/
	//	printf("*** %d %d %d \n", O_RDONLY, O_WRONLY, O_RDWR);
	fd_t = open(buf, as->r[17].ld(), as->r[18].ld());/* open() here */
      }
      else{
	buf[1]='S';
	buf[2]='A';
	int flag = as->r[17].ld() | O_CREAT;
	fd_t = open(buf, flag, as->r[18].ld()); /* open() here */
      }

      /* mapping from the real to the logical */
      if(fd_t==-1){
	ret = -1;
	as->r[0].init(errno);
	as->r[19].init(~0ull);
      }
      else{
	ret = fd_table[fd_t];
	as->r[0].init(ret);
	as->r[19].init(0);
      }

      if(F) printf("%8qd: open(\"%s\", 0x%qx, 0x%qx) = logic %qd(real %d)\n",
		   e->retired_inst, 
		   buf, as->r[17].ld(), as->r[18].ld(),
		   ret, fd_t);
    }
    break;
    /************************************************************/
  case ASYS_sigreturn: /*** Note that sigreturn updates as->pc ***/
    {
      uint64_t pt[OSF_SIGCONTEXT_SIZE/8];
      char *buf = (char *)pt;

      data_t d;
      data_t adr;
      for(int i=0; i<OSF_SIGCONTEXT_SIZE; i++){
	adr.init(as->r[16].ld() + i);
	mem->ld_nbyte(1, &adr, &d);
        buf[i] = d.ld();
      }
      for(int i=0; i<31; i++) as->r[i].init(pt[i+4]);
      for(int i=0; i<31; i++) as->f[i].init(pt[i+37]);
      as->pc.init(pt[2]);  /** Update PC **/

      if(F) printf("%8qd: sigreturn(0x%qx), next pc:0x%qx\n", 
		   e->retired_inst, as->r[16].ld(), as->pc.ld());
      break;
    }
    
    /************************************************************/
  case ASYS_close:
    {
      int fd_t = translate_fd(as->r[16].ld());

      if(fd_t!=0 && fd_t!=1 && fd_t!=2){
	ret = close(fd_t);
	as->r[0].init(ret);
      }
      as->r[19].init(0);
      if(F) printf("%8qd: close(%qd(%d)) = %qd\n", 
		   e->retired_inst, as->r[16].ld(),
		   fd_t, ret);
      break;
    }
  case ASYS_getpagesize:
    {
      ret = 8192; /* getpagesize(); */
      as->r[0].init(ret);
      as->r[19].init(0);
      if(F) printf("%8qd: getpagesize() = 0x%llx\n", 
		   e->retired_inst, ret);
      break;
    }

    /************************************************************/        
  case ASYS_sigprocmask:
    ret = 0;
    as->r[0].init(ret);
    as->r[19].init(ret);
    if(F) printf("%8qd: sigprocmask(%qd,%qx) = %qd\n", 
		 e->retired_inst, as->r[16].ld(), 
		 as->r[17].ld(), ret);
    break;

    /************************************************************/        
  case ASYS_getrusage: // Not Checked 
    {
      struct rusage ru;
      ret = getrusage((enum __rusage_who)as->r[16].ld(), &ru);

      for(int i=0; i<OSF_RUSAGE_SIZE; i++){
	data_t adr; adr.init(as->r[17].ld()+i);
	data_t dat; dat.init(0);
	mem->st_nbyte(1, &adr, &dat);
      }
      
      as->r[0].init(ret);
      as->r[19].init(0);

      if(F) printf("%8qd: getrusage(%qd, %qx) = %qd\n", 
		   e->retired_inst, 
		   as->r[16].ld(), as->r[17].ld(), ret);
      break;
    }

    /************************************************************/    
  case 	ASYS_gettimeofday:
    {
      as->r[19].init(0);
      if(F) printf("%8qd: gettimeofday(%qd,%qx) = %qd\n", 
		   e->retired_inst, as->r[16].ld(), 
		   as->r[17].ld(), ret);
      break;
    }

    /************************************************************/        
  case ASYS_fcntl:
    {
      ret = 0;
      as->r[0].init(ret);
      as->r[19].init(0);
      if(F) printf("%8qd: fcntl(%qd,%qx) = %qd\n", 
		   e->retired_inst, as->r[16].ld(), 
		   as->r[17].ld(), ret);
      break;
    }

    /************************************************************/    
  case ASYS_sigaction:
    {
      ret = 0;
      as->r[0].init(ret);
      as->r[18].init(as->r[17].ld());
      as->r[19].init(0);
      if(F) printf("%8qd: sigaction(%qx,%qx,%qx,%qx) = %qd\n", 
		   e->retired_inst,
		   as->r[16].ld(), as->r[17].ld(),
		   as->r[18].ld(), as->r[19].ld(), ret);
      break;
    }

    /************************************************************/    
  case ASYS_osf_getsysinfo:
    {
      ret = 0;
      as->r[0].init(ret);
      as->r[19].init(0);
      if(F) printf("%8qd: osf_getsysinfo(%qx,%qx,%qx,%qx) = %qd\n", 
		   e->retired_inst, 
		   as->r[16].ld(), as->r[17].ld(),
		   as->r[18].ld(), as->r[19].ld(), ret);
      break;
    }

    /************************************************************/    
  case ASYS_osf_setsysinfo:
    {
      ret = 0;
      as->r[0].init(ret);
      //    as->r[20]= ret;
      as->r[19].init(0);
      if(F) printf("%8qd: osf_setsysinfo(%qx,%qx,%qx,%qx) = %qd\n", 
		   e->retired_inst, 
		   as->r[16].ld(), as->r[17].ld(),
		   as->r[18].ld(), as->r[19].ld(), ret);
      break;
    }

    /************************************************************/    
  case ASYS_getuid:
    {
      ret = getuid();
      as->r[0].init(ret);
      as->r[19].init(0);
      as->r[20].init(geteuid());
      if(F) printf("%8qd: getuid() = %qd\n", e->retired_inst, ret);
      break;
    }

    /************************************************************/    
  case ASYS_getgid:
    {
      ret = getgid();
      as->r[0].init(ret);
      as->r[19].init(0);
      as->r[20].init(getegid());
      if(F) printf("%8qd: getgid(0x%qx) = %qd\n", e->retired_inst, 
		   as->r[16].ld(), ret);
      break;
    }

    /************************************************************/    
  case ASYS_brk:
    {
      ret = as->r[16].ld();
      as->r[0].init(ret);
      as->r[19].init(0);
      if(F) printf("%8qd: brk(0x%qx) = %qx\n", e->retired_inst, 
		   as->r[16].ld(), ret);
      break;
    }

    /************************************************************/    
  case ASYS_getpid:
    {
      ret = getpid();
      as->r[0].init(ret);
      as->r[19].init(0);
      if(F) printf("%8qd: getpid() = %qx\n", 
		   e->retired_inst, ret);
      break;
    }
    /************************************************************/    
    /*
  case ASYS_stat: 
    {
      if(F) printf("%8qd: stat() = %qx\n", 
		   e->retired_inst, ret);
      break;
    }
    */
    /************************************************************/    
  case ASYS_fstat:
    {
      int fd_t = translate_fd(as->r[16].ld());

      struct stat sbuf;
      ret = fstat(fd_t,  &sbuf);

      struct osf_statbuf asbuf;
      memset(&asbuf, 0, sizeof(struct osf_statbuf));
      asbuf.st_dev     = sbuf.st_dev;
      asbuf.st_ino     = sbuf.st_ino;
      asbuf.st_mode    = sbuf.st_mode;
      asbuf.st_nlink   = sbuf.st_nlink;
      asbuf.st_uid     = sbuf.st_uid;
      asbuf.st_gid     = sbuf.st_gid;
      asbuf.st_rdev    = sbuf.st_rdev;
      asbuf.st_size    = sbuf.st_size;
      asbuf.st_atime   = sbuf.st_atime;
      asbuf.st_mtime   = sbuf.st_mtime;
      asbuf.st_ctime   = sbuf.st_ctime;
      asbuf.st_blksize = sbuf.st_blksize;
      asbuf.st_blocks  = sbuf.st_blocks;

      char *buf = (char *)&asbuf;

      for(int i=0; i<OSF_FSTAT_SIZE; i++){
	data_t adr; 
	adr.init(as->r[17].ld()+i);
	data_t dat; 
	dat.init(buf[i]);
	mem->st_nbyte(1, &adr, &dat);
      }
      
      as->r[0].init(ret);
      as->r[19].init(0);
      if(F) printf("%8qd: fstat(logic %qd(real %d), 0x%qx) = %qd\n", 
		   e->retired_inst, as->r[16].ld(), 
		   fd_t, as->r[17].ld(), ret);
      break; 
    }
    /************************************************************/    
  case ASYS_ioctl: /** now write 0, should modify **/
    {
      int fd_t = translate_fd(as->r[16].ld());

      switch(as->r[17].ld()){
      case 0x40067408:      // OSF_TIOCGRETP 
	{
	  struct osf_sgttyb {
	    uint8_t  sg_ispeed;     /* input speed */
	    uint8_t  sg_ospeed;     /* output speed */
	    uint8_t  sg_erase;      /* erase character */
	    uint8_t  sg_kill;       /* kill character */
	    uint16_t sg_flags;      /* mode flags */
	  };
	  
	  struct termios lbuf;
	  ret = tcgetattr(fd_t, &lbuf);

	  /*** store 6 byte, value of 0 ***/
	  for(int i=0; i<6; i++){
	    data_t adr; adr.init(as->r[18].ld()+i);
	    data_t dat; dat.init(0);
	    mem->st_nbyte(1, &adr, &dat);
	  }

	  if(ret!=-1){
	    as->r[0].init(ret);
	    as->r[19].init(0); // A3
	  }
	  else{
	    as->r[0].init(errno);
	    as->r[19].init(ret);
	  }
	  break;
	}

	
      default:
	ret = 0;
	as->r[19].init(ret);
	if(F) printf("*** ioctl default\n");
	break;
      }

      if(F)
	printf("%8qd: ioctl(logic %qd(real %d), 0x%qx, 0x%qx) = 0x%qx\n", 
	       e->retired_inst, as->r[16].ld(), fd_t,
	       as->r[17].ld(), as->r[18].ld(), ret); 
      break;
    }

    /************************************************************/    
  case ASYS_read:
    {
      char *buf = (char *)calloc(1, as->r[18].ld());
      if(buf==NULL) printf("*** malloc error in syscall.cc::read\n");

      int fd_t = translate_fd(as->r[16].ld());
      if(ms->fd_stdin_flag && fd_t==0) fd_t=ms->fd_stdin; /* STDIN */
      
      ret = read(fd_t, buf, as->r[18].ld());
      
      /*      for(int i=0; i<ret; i++){  */  // This is better.
      for(int i=0; i<(int)as->r[18].ld(); i++){ // simple scalar imple!
	DATA_TYPE d = buf[i];
	data_t adr; adr.init(as->r[17].ld()+i);
	data_t dat; dat.init(d);
	mem->st_nbyte(1, &adr, &dat);
      }
      as->r[0].init(ret);
      as->r[19].init(0);
      free(buf);

      if(F) printf("%8qd: read(logic %qd(real %d),%qx,0x%qx) = 0x%qx\n", 
		   e->retired_inst, as->r[16].ld(),
		   fd_t, as->r[17].ld(), as->r[18].ld(), 
		   ret);
      break;
    }
    
    /************************************************************/
  case ASYS_write:
    {
      char *buf = (char *)malloc(as->r[18].ld());
      if(buf==NULL) printf("*** malloc error in syscall.cc::write\n");
	
      for(int i=0; i<(int)as->r[18].ld(); i++){
	data_t d;
	data_t adr;
	adr.init(as->r[17].ld()+i);
	mem->ld_nbyte(1, &adr, &d);
	buf[i] = d.ld();
      }

      int fd_t = translate_fd(as->r[16].ld());

      ret = write(fd_t, (void *)buf, (long)as->r[18].ld());
      as->r[0].init(ret);
      as->r[19].init(0);
      free(buf);

      if(F) printf("%8qd: write(%qd(%d), 0x%qx, %qd) = %qd\n", 
		   e->retired_inst, as->r[16].ld(),
		   fd_t, as->r[17].ld(), as->r[18].ld(), ret);
      break;
    }


    /************************************************************/
  case ASYS_lseek:
    {
      int fd_t = translate_fd(as->r[16].ld());
      if(ms->fd_stdin_flag && fd_t==0) fd_t=ms->fd_stdin; /* STDIN */

      ret = lseek(fd_t, as->r[17].ld(), as->r[18].ld());
      if(ret==-1){
	as->r[0].init(errno);
	as->r[19].init(~0);
      }
      else{
	as->r[0].init(ret);
	as->r[19].init(0);
      }
      if(F) printf("%8qd: lseek(logic %qd(real %d), %qd, %qd) = 0x%qx\n",
		   e->retired_inst, 
		   as->r[16].ld(), fd_t, 
		   as->r[17].ld(), as->r[18].ld(), ret);
      break;
    }
    
    /************************************************************/
  case ASYS_getrlimit:
    {
      ret = 0;
      as->r[19].init(ret);
      if(F) printf("%8qd: getrlimit(%qd,%qd,%qd) = %qd\n",
		   e->retired_inst, as->r[16].ld(), 
		   as->r[17].ld(), as->r[18].ld(), ret);
      break;
    }

    /************************************************************/
  case ASYS_setrlimit:
    {
      ret = 0;
      as->r[19].init(ret);
      if(F) printf("%8qd: setrlimit(%qd,%qd,%qd) = %qd\n",
		   e->retired_inst, as->r[16].ld(), 
		   as->r[17].ld(), as->r[18].ld(), ret);
      break;
    }

    /************************************************************/
  case ASYS_exit:
    {
      ret = 0;
      running = 0;
      if(F) printf("%8qd: exit() = %qd\n", e->retired_inst, ret);
      break;
    }    

  default:
    {
      printf("TC: %qd systemcall not implemented!!: %qd\n" ,
	     e->retired_inst, as->r[0].ld());
      as->r[0].init(0);
      as->r[19].init((uint64_t)-1);
    }
  }
  
  fflush(stdout);
}
